
import logging


class LineParseException(Exception):
    pass


class MissingDefaultException(Exception):
    def __init__(self, task):
        self.task = task


class BaseReader(object):
    """
    Abstract reader class
    """
    def __init__(self, settings, fpath):
        self.fpath = fpath
        self.tasks = tuple(self.get_tasks())
        self.tasks_defaults = {task['name']: task.get("default")
                               for task in settings.tasks if task['name'] in self.tasks}

    def get_default(self, task, value):
        """
        Get default value for task if given
        """
        default = self.tasks_defaults.get(task)
        if default is not None:
            if default.lower() == 'copy':
                return value
            return default

        raise MissingDefaultException(task)

    def readsents(self, silent=True, only_tokens=False):
        """
        Generator over dataset sentences. Each output is a tuple of (Input, Tasks)
        objects with, where each entry is a list of strings.
        """
        current_sent, lines = 0, self.parselines()

        while True:
            try:
                line = next(lines)
                current_sent += 1
                # check if parse exception
                if isinstance(line, LineParseException):
                    if not silent:
                        logging.warning(
                            "Parse error at [{}:sent={}]\n  => {}"
                            .format(self.fpath, current_sent + 1, str(line)))
                    continue

                inp, tasks = line
                if only_tokens:
                    yield inp
                else:
                    yield (self.fpath, current_sent), (inp, tasks)

            except StopIteration:
                break

    def check_tasks(self, expected=None):
        """
        Returns the union of the tasks across files. If `expected` is passed,
        it is understood to be a subset of all available tasks and it will check
        whether all `expected` are in fact available in all files as per the output
        of `get_tasks`.

        Returns
        ========
        List of tasks available in all files (or a subset if `expected` is passed)
        """
        if expected is not None:
            diff = set(expected).difference(set(self.tasks))
            if diff:
                raise ValueError("Following expected tasks are missing "
                                 "from at least one file: '{}'"
                                 .format('"'.join(diff)))

        return tuple(expected or self.tasks)

    def parselines(self):
        """
        Generator over tuples of (Input, Tasks) holding input and target data
        as lists of strings. Some tasks might be missing from a file, in which
        case the corresponding value is None.

        ParseError can be thrown if an issue is encountered during processing.
        The issue can be documented by passing an error message as second argument
        to ParseError
        """
        raise NotImplementedError

    def get_tasks(self):
        """
        Reader is responsible for extracting the expected tasks in the file.

        Returns
        =========
        Set of tasks in a file
        """
        raise NotImplementedError



# Can be run with python -m pie.scripts.evaluate
import os.path
from pie import utils


from pie.models import BaseModel
from pie.data import Dataset, Reader
from pie.settings import load_default_settings, settings_from_file


def run(model_path, test_path, train_path, # data
        # decoding
        settings, batch_size, buffer_size, use_beam, beam_width, device, model_info,
        full, confusion, report, markdown):  # report

    model = BaseModel.load(model_path).to(device)
    if model_info:
        print(model)

    # settings
    if hasattr(model, '_settings'):  # new models should all have _settings
        settings = model._settings
    elif settings:
        print("Using user specified settings file: {}".format(settings))
        with utils.shutup():
            settings = settings_from_file(settings)
    else:
        print("Warning! Using default settings")
        with utils.shutup():
            settings = load_default_settings()

    # overwrite defaults
    settings.batch_size = batch_size
    settings.buffer_size = buffer_size
    settings.device = device
    settings.shuffle = False    # avoid shuffling

    # read datasets
    trainset = None
    if train_path:
        trainset = Dataset(
            settings, Reader(settings, train_path), model.label_encoder)
    elif hasattr(settings, "input_path") and \
         settings.input_path and os.path.exists(settings.input_path):
        print("--- Using train set from settings")
        trainset = Dataset(
            settings, Reader(settings, settings.input_path), model.label_encoder)
    else:
        print("--- Not using trainset to evaluate known/unknown tokens")

    if not len(test_path) and hasattr(settings, "test_path"):
        print("--- Using test set from settings")
        test_path = (settings.test_path, )

    testset = Dataset(settings, Reader(settings, *test_path), model.label_encoder)

    for task in model.evaluate(testset, trainset,
                               use_beam=use_beam, beam_width=beam_width).values():
        task.print_summary(
            full=full, confusion_matrix=confusion, report=report, markdown=markdown)



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('model_path')
    parser.add_argument('test_path', help="unix string", nargs='+')
    parser.add_argument('--train_path',
                        help="needed to compute ambiguous and unknown tokens")
    parser.add_argument('--settings', help="settings file used for training")
    parser.add_argument('--batch_size', type=int, default=500)
    parser.add_argument('--buffer_size', type=int, default=100000)
    parser.add_argument('--use_beam', action='store_true')
    parser.add_argument('--beam_width', type=int, default=12)
    parser.add_argument('--device', default='cpu')
    parser.add_argument('--model_info', action='store_true')
    parser.add_argument('--full', action='store_true')
    parser.add_argument('--confusion', default=False, action="store_true")
    parser.add_argument('--report', default=False,
                        action="store_true", help="Get full report on each class")
    parser.add_argument('--markdown', default=False,
                        action="store_true", help="Use Markdown")
    args = parser.parse_args()
    run(model_path=args.model_path, test_path=args.test_path,
        train_path=args.train_path, settings=args.settings,
        batch_size=args.batch_size, buffer_size=args.buffer_size,
        use_beam=args.use_beam, beam_width=args.beam_width,
        device=args.device, model_info=args.model_info,
        full=args.full, confusion=args.confusion, report=args.report,
        markdown=args.markdown)




import click
# Can be run with python -m pie.scripts.group
import pie.utils


@click.group()
def pie_cli():
    """ Group command for Pie """


@pie_cli.command()
@click.option('--model_spec', help="Path to model(s)")
@click.option('--batch_size', type=int, default=50, help="Size of the batch")
@click.option('--device', default='cpu', help="Device to use to run the network")
def webapp(model_spec, batch_size, device):
    """ Run the webapp """
    # Until further version, we should explain what's going on
    print("Not supported anymore, do pip install flask_pie")
    raise Exception("The web version of pie has moved to github.com/"
                    "hipster-philology/flask_pie")


@pie_cli.command()
@click.argument('input_path')
@click.argument('model_spec', type=pie.utils.model_spec)
@click.option('--keep_boundaries', is_flag=True,
              help='Keep boundaries from the original input file')
@click.option('--batch_size', type=int, default=50)
@click.option('--device', default='cpu')
@click.option('--use_beam', is_flag=True, default=False)
@click.option('--beam_width', default=10, type=int)
@click.option('--lower', is_flag=True, help="Treat the input as lower case")
@click.option('--max_sent_len', type=int, default=35,
              help='Split sentences longer than this amount')
@click.option('--vrt', is_flag=True, help='Verticalized input format')
def tag(model_spec, input_path, keep_boundaries, batch_size, device,
        use_beam, beam_width, lower, max_sent_len, vrt):
    """ Tag [INPUT_PATH] with model(s) at [MODEL_SPEC]"""
    import pie.scripts.tag
    pie.scripts.tag.run(
        model_spec, input_path, beam_width, use_beam, keep_boundaries,
        device=device, batch_size=batch_size, lower=lower,
        max_sent_len=max_sent_len, vrt=vrt)


@pie_cli.command("tag-pipe")
@click.argument('model_spec', type=pie.utils.model_spec)
@click.option('--batch_size', type=int, default=50)
@click.option('--device', default='cpu')
@click.option('--use_beam', is_flag=True, default=False)
@click.option('--beam_width', default=10, type=int)
@click.option('--lower', is_flag=True, help="Lowercase input to tagger", default=False)
@click.option('--tokenize', is_flag=True, help="Tokenize the input", default=False)
def tag_pipe(model_spec, device, batch_size, lower, beam_width, use_beam, tokenize):
    """ Tag the terminal input with [MODEL_SPEC]"""
    import pie.scripts.tag_pipe
    pie.scripts.tag_pipe.run(
        model_spec=model_spec, device=device, batch_size=batch_size,
        lower=lower, beam_width=beam_width, use_beam=use_beam, tokenize=tokenize)


@pie_cli.command("eval")
@click.argument('model_path')
@click.argument('test_path', nargs=-1)
@click.option('--train_path', default=None,
              help="File used to compute unknown tokens/targets")
@click.option('--settings', help="Settings file used for training")
@click.option('--batch_size', type=int, default=500)
@click.option('--buffer_size', type=int, default=100000)
@click.option('--device', default='cpu')
@click.option('--model_info', is_flag=True, default=False)
@click.option('--full', is_flag=True, default=False)
@click.option('--confusion', default=False, is_flag=True)
@click.option('--report', default=False, is_flag=True)
@click.option('--markdown', default=False, is_flag=True)
@click.option('--use_beam', is_flag=True, default=False)
@click.option('--beam_width', type=int, default=12)
def evaluate(model_path, test_path, train_path, settings, batch_size,
             buffer_size, device, model_info, full, confusion, report,
             markdown, use_beam, beam_width):
    """ Evaluate [MODEL_PATH] against [TEST_PATH] using [TRAIN_PATH] to compute
    unknown tokens"""
    import pie.scripts.evaluate
    pie.scripts.evaluate.run(
        model_path=model_path, test_path=test_path, train_path=train_path,
        settings=settings, batch_size=batch_size, buffer_size=buffer_size,
        device=device, model_info=model_info, full=full, confusion=confusion,
        report=report, markdown=markdown,
        use_beam=use_beam, beam_width=beam_width)


@pie_cli.command("train")
@click.argument('config_path')
def train(config_path):
    """ Train a model using the file at [CONFIG_PATH]"""
    import pie.scripts.train
    import pie.settings
    pie.scripts.train.run(pie.settings.settings_from_file(config_path))


@pie_cli.command("info")
@click.argument("model_file", type=click.Path(exists=True,
                                              file_okay=True,
                                              dir_okay=False,
                                              readable=True))
def info(model_file):
    from pie.models import BaseModel
    import pprint
    m = BaseModel.load(model_file)
    bar = "=====================\n"
    click.echo(bar+"Settings", color="red")
    pprint.pprint(m._settings)
    click.echo(bar+"Architecture", color="red")
    click.echo(repr(m))


if __name__ == "__main__":
    pie_cli()




import torch
import torch.nn as nn
import torch.nn.functional as F

from pie import initialization, torch_utils


def DotScorer(dec_out, enc_outs, **kwargs):
    """
    Score for query decoder state and the ith encoder state is given
    by their dot product.

    dec_outs: (trg_seq_len x batch x hid_dim)
    enc_outs: (src_seq_len x batch x hid_dim)

    output: ((trg_seq_len x) batch x src_seq_len)
    """
    score = torch.bmm(
        # (batch x src_seq_len x hid_dim)
        enc_outs.transpose(0, 1),
        # (batch x hid_dim x trg_seq_len)
        dec_out.transpose(0, 1).transpose(1, 2))
    # (batch x src_seq_len x trg_seq_len) => (trg_seq_len x batch x src_seq_len)
    return score.transpose(0, 1).transpose(0, 2)


class GeneralScorer(nn.Module):
    """
    Inserts a linear projection to the query state before the dot product
    """
    def __init__(self, dim):
        super(GeneralScorer, self).__init__()
        self.W_a = nn.Linear(dim, dim, bias=False)
        self.init()

    def init(self):
        initialization.init_linear(self.W_a)

    def forward(self, dec_out, enc_outs, **kwargs):
        return DotScorer(self.W_a(dec_out), enc_outs)


class BahdanauScorer(nn.Module):
    """
    Projects both query decoder state and encoder states to an attention space.
    The scores are computed by a dot product with a learnable param v_a after
    transforming the sum of query decoder state and encoder state with a tanh.

    `score(a_i_j) = a_v \dot tanh(W_s @ h_s_j + W_t @ h_t_i)`
    """
    def __init__(self, hid_dim, att_dim):
        self.att_dim = att_dim
        super(BahdanauScorer, self).__init__()
        # params
        self.W_s = nn.Linear(hid_dim, att_dim, bias=False)
        self.W_t = nn.Linear(hid_dim, att_dim, bias=True)
        self.v_a = nn.Parameter(torch.Tensor(att_dim, 1))
        self.init()

    def init(self):
        torch.nn.init.uniform_(self.v_a, -0.05, 0.05)
        initialization.init_linear(self.W_s)
        initialization.init_linear(self.W_t)

    def project_enc_outs(self, enc_outs):
        """
        mapping: (seq_len x batch x hid_dim) -> (seq_len x batch x att_dim)

        Returns:
        --------
        torch.Tensor (seq_len x batch x att_dim),
            Projection of encoder output onto attention space
        """
        seq_len, batch, hid_dim = enc_outs.size()

        return self.W_s(enc_outs.view(-1, hid_dim)).view(seq_len, batch, -1)

    def forward(self, dec_out, enc_outs, enc_att=None):
        # get projected encoder sequence
        if enc_att is None:
            # (seq_len x batch x att dim)
            enc_att = self.project_enc_outs(enc_outs)

        # get projected decoder step (or sequence in ffw mode)
        # ((trg_len x) batch x att_dim)
        dec_att = self.W_t(dec_out)

        (src_len, batch, dim), (trg_len, *_) = enc_att.size(), dec_att.size()

        # see seqmod/test/modules/attention.py
        enc_att = enc_att.view(src_len, 1, batch, -1)
        # (src_len x trg_len x batch x att_dim) +
        # (1           x trg_len x batch x att_dim)
        # => (src_len x trg_len x batch x att_dim)
        dec_enc_att = torch.tanh(enc_att + dec_att.unsqueeze(0))

        # (src_len * trg_len x batch x dim)
        dec_enc_att = dec_enc_att.view(src_len * trg_len, batch, dim)
        # (batch x src_len * trg_len x dim)
        dec_enc_att = dec_enc_att.transpose(0, 1).contiguous()
        # (batch x seq_len x dim) * (1 x dim x 1) -> (batch x seq_len)
        scores = dec_enc_att @ self.v_a.unsqueeze(0)
        # (batch x src_len x trg_len)
        scores = scores.view(batch, src_len, trg_len)
        # (trg_len x batch x src_len)
        return scores.transpose(0, 2).transpose(1, 2)


class Attention(nn.Module):
    """
    Global attention implementing the three scorer modules from Luong 15.

    Parameters:
    -----------
    - hid_dim: int, dimensionality of the query vector
    - att_dim: (optional) int, dimensionality of the attention space (only
        used by the bahdanau scorer). If not given it will default to hid_dim.
    - scorer: str, one of ('dot', 'general', 'bahdanau')
    - hid_dim2: (optional), int, dimensionality of the key vectors (optionally
        used by the bahdanau scorer if given)
    """
    def __init__(self, hid_dim, att_dim=None, scorer='general'):
        super(Attention, self).__init__()

        # Scorer
        if scorer.lower() == 'dot':
            self.scorer = DotScorer
        elif scorer.lower() == 'general':
            self.scorer = GeneralScorer(hid_dim)
        elif scorer.lower() == 'bahdanau':
            self.scorer = BahdanauScorer(hid_dim, att_dim or hid_dim)

        # Output layer (Luong 15. eq (5))
        self.linear_out = nn.Linear(
            hid_dim * 2, hid_dim, bias=scorer.lower() == 'bahdanau')

        self.init()

    def init(self):
        initialization.init_linear(self.linear_out)

    def forward(self, dec_out, enc_outs, lengths):
        """
        Parameters:
        -----------

        - dec_outs: torch.Tensor(trg_seq_len x batch_size x hid_dim)
        - enc_outs: torch.Tensor(seq_len x batch_size x hid_dim)
        - lengths: torch.LongTensor(batch), source lengths

        Returns:
        --------
        - context: (trg_seq_len x batch x hid_dim)
        - weights: (trg_seq_len x batch x seq_len)
        """
        # get scores
        # (trg_seq_len x batch x seq_len)
        weights = self.scorer(dec_out, enc_outs)

        # apply source length mask
        mask = torch_utils.make_length_mask(lengths)
        # (batch x src_seq_len) => (trg_seq_len x batch x src_seq_len)
        mask = mask.unsqueeze(0).expand_as(weights)
        # weights = weights * mask.float()
        # Torch 1.1 -> 1.2: (1 - mask) becomes ~(mask)
        weights.masked_fill_(~mask, -float('inf'))

        # normalize
        weights = F.softmax(weights, dim=2)

        # (eq 7) (batch x trg_seq_len x seq_len) * (batch x seq_len x hid_dim)
        # => (batch x trg_seq_len x hid_dim) => (trg_seq_len x batch x hid_dim)
        context = torch.bmm(
            weights.transpose(0, 1), enc_outs.transpose(0, 1)
        ).transpose(0, 1)
        # (eq 5) linear out combining context and hidden
        context = torch.tanh(self.linear_out(torch.cat([context, dec_out], 2)))

        return context, weights




import collections
from difflib import SequenceMatcher


def segment(s, infix):
    idx = s.find(infix)
    return s[:idx], s[idx+len(infix):]


def get_segments(a, b, match):
    infix = a[match.a: match.a + match.size]
    return segment(a, infix), segment(b, infix)


def lcs(a, b):
    return SequenceMatcher(None, a, b).find_longest_match(0, len(a), 0, len(b))


def breadth_first(root):
    def breadth_first_(level, *nodes):
        children = []
        for node in nodes:
            yield level, node
            children.extend(node.children())
        if children:
            yield from breadth_first_(level+1, *children)

    return list(breadth_first_(1, root))


def depth_first(root):
    def depth_first_(level, node):
        yield level, node
        for child in node.children():
            yield from depth_first_(level+1, child)

    return list(depth_first_(1, root))


class Node:
    def __init__(self, _prefix, _infix, _suffix, target):
        self._prefix = _prefix  # prefix of the string at this level
        self._infix = _infix    # current string
        self._suffix = _suffix  # suffix of the string at this level
        self.target = target
        # children
        self.prefix = None
        self.suffix = None

    def __len__(self):
        return len(self._infix)

    def __repr__(self):
        tree = ""
        for level, child in depth_first(self):
            tree += ' ' * (level - 1) * 2
            if isinstance(child, Leaf):
                tree += '- {}\n'.format(child)
            else:
                tree += '{}({},{})\n'.format(child._infix, *child.get_span())
        return tree

    def get_source(self):
        return self._prefix + self._infix + self._suffix

    def get_span(self):
        return len(self._prefix), len(self._suffix)

    def get_depth(self):
        return max(1 + self.prefix.get_depth() if self.prefix else 0,
                   1 + self.suffix.get_depth() if self.suffix else 0)

    def children(self):
        return [self.prefix, self.suffix]

    def to_class(self):
        return get_class(self)

    def to_tuple(self):
        if isinstance(self, Leaf):
            # full replacement
            return (len(self.a), self.b, 0, "")
        else:
            plen, slen = self.get_span()
            start = self.target.find(self._infix)
            pstring, sstring = self.target[:start], self.target[start+len(self._infix):]
            return (plen, pstring, slen, sstring)


class Leaf(Node):
    def __init__(self, a, b):
        self.a = a
        self.b = b
        # cache rule and string representation
        self.string = self.rule = None

        if a and b:
            self.rule = Rule('replace', None, b)
            self.string = '<replace({})>'.format(b)
        elif a:
            self.rule = Rule('delete', a, None)
            self.string = '<delete({})>'.format(a)
        elif b:
            self.rule = Rule('insert', None, b)
            self.string = '<insert({})>'.format(b)
        else:
            self.rule = Rule('keep', None, None)
            self.string = '<keep>'

    def __repr__(self):
        return self.string

    def __len__(self):
        return 0

    def get_depth(self):
        return 1

    def children(self):
        return []


Rule = collections.namedtuple('Rule', ['action', 'a', 'b'])


def apply_rule(rule, inp, prefix=False):
    if rule.action == 'replace':
        return rule.b
    elif rule.action == 'delete':
        return inp[len(rule.a):] if prefix else inp[:-len(rule.a)]
    elif rule.action == 'insert':
        return (rule.b + inp) if prefix else (inp + rule.b)
    else:
        return inp


def get_class(node):
    if isinstance(node, Leaf):
        return node.rule
    else:
        return node.get_span(), (get_class(node.prefix), get_class(node.suffix))


def make_edit_tree(a, b):
    match = lcs(a, b)

    # check if leaf
    if match.size == 0:
        return Leaf(a, b)

    # parent
    else:
        (pre_a, suf_a), (pre_b, suf_b) = get_segments(a, b, match)
        node = Node(pre_a, a[match.a:match.a+match.size], suf_a, b)
        node.prefix = make_edit_tree(pre_a, pre_b)
        node.suffix = make_edit_tree(suf_a, suf_b)
        return node


def apply_edit_tree(tclass, inp, prefix=False):
    if isinstance(tclass, Rule):
        # apply leaf
        return apply_rule(tclass, inp, prefix)
    else:
        # split and recur on non-terminals
        (plen, slen), (p, s) = tclass
        infix = inp[plen:len(inp)-slen]
        prefix = apply_edit_tree(p, inp[:plen], prefix=True)
        suffix = apply_edit_tree(s, inp[len(inp)-slen:])
        return prefix + infix + suffix


def apply_tuple(tup, inp):
    plen, pstring, slen, sstring = tup
    return pstring + inp[plen:-slen] + sstring


def transform(lem, tok):
    return make_edit_tree(tok, lem).to_class()


def inverse_transform(pred, tok):
    return apply_edit_tree(pred, tok)




# Can be run with python -m pie.scripts.tag
from pie import utils
from pie.tagger import Tagger


def run(model_spec, input_path, beam_width, use_beam, keep_boundaries, **kwargs):
    tagger = Tagger(**kwargs)

    for model, tasks in model_spec:
        tagger.add_model(model, *tasks)
        print(" - model: {}".format(model))
        tasks = tasks or tagger.models[-1][0].label_encoder.tasks
        print(" - tasks: {}".format(", ".join(tasks)))

    for fpath in utils.get_filenames(input_path):
        print("Tagging file [{}]...".format(fpath))
        tagger.tag_file(fpath, use_beam=use_beam, beam_width=beam_width,
                        keep_boundaries=keep_boundaries)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('model_spec', type=utils.model_spec)
    parser.add_argument('input_path', help="unix string")
    parser.add_argument('--keep_boundaries', action='store_true')
    parser.add_argument('--batch_size', type=int, default=50)
    parser.add_argument('--device', default='cpu')
    parser.add_argument('--use_beam', action='store_true')
    parser.add_argument('--beam_width', default=10, type=int)
    parser.add_argument('--lower', action='store_true')
    parser.add_argument('--max_sent_len', type=int, default=35)
    parser.add_argument('--vrt', action='store_true')
    args = parser.parse_args()

    run(args.model_spec, args.input_path,
        beam_width=args.beam_width, use_beam=args.use_beam,
        keep_boundaries=args.keep_boundaries,
        device=args.device, batch_size=args.batch_size,
        lower=args.lower, max_sent_len=args.max_sent_len,
        vrt=args.vrt)




import math

import torch
import torch.nn as nn
from pie import torch_utils


class CustomLSTMCell(nn.Module):
    """A basic LSTM cell."""

    def __init__(self, input_size, hidden_size):
        """
        Most parts are copied from torch.nn.LSTMCell.
        """
        super(CustomLSTMCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = nn.Parameter(torch.Tensor(input_size, 4 * hidden_size))
        self.weight_hh = nn.Parameter(torch.Tensor(hidden_size, 4 * hidden_size))
        self.bias = nn.Parameter(torch.FloatTensor(4 * hidden_size))
        self.reset_parameters()

    def reset_parameters(self):
        """
        Initialization from dynet
        """
        nn.init.constant_(self.bias, 0)
        nn.init.constant_(self.bias[self.hidden_size: 2 * self.hidden_size], 1.0)
        nn.init.xavier_uniform_(self.weight_hh)
        nn.init.xavier_uniform_(self.weight_ih)

    def forward(self, input_, hx):
        """
        Args:
            input_: A (batch, input_size) tensor containing input
                features.
            hx: A tuple (h_0, c_0), which contains the initial hidden
                and cell state, where the size of both states is
                (batch, hidden_size).

        Returns:
            h_1, c_1: Tensors containing the next hidden and cell state.
        """
        h_0, c_0 = hx
        batch_size = h_0.size(0)
        bias = self.bias.unsqueeze(0).expand(batch_size, *self.bias.size())
        wi = torch.mm(input_, self.weight_ih)
        wh = torch.mm(h_0, self.weight_hh)
        i, f, o, g = torch.split(wh + wi + bias, self.hidden_size, dim=1)
        c_1 = torch.sigmoid(f + 1) * c_0 + torch.sigmoid(i) * torch.tanh(g)
        h_1 = torch.sigmoid(o) * torch.tanh(c_1)
        return h_1, c_1

    def __repr__(self):
        s = '{name}({input_size}, {hidden_size})'
        return s.format(name=self.__class__.__name__, **self.__dict__)


class CustomLSTM(nn.Module):
    """A module that runs multiple steps of LSTM."""
    def __init__(self, input_size, hidden_size, num_layers=1, **kwargs):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cell = CustomLSTMCell(input_size=input_size, hidden_size=hidden_size)

    @staticmethod
    def _forward_rnn(cell, input_, hx, lengths, backward=False):
        max_time = input_.size(0)
        output = []
        for time in reversed(range(max_time)) if backward else range(max_time):
            h_next, c_next = cell(input_=input_[time], hx=hx)
            mask = (time < lengths).float().unsqueeze(1).expand_as(h_next)
            h_next = h_next * mask + hx[0] * (1 - mask)
            c_next = c_next * mask + hx[1] * (1 - mask)
            hx_next = (h_next, c_next)
            output.append(h_next * mask)
            hx = hx_next
        output = torch.stack(output, 0)
        return output, hx

    def forward(self, input_, hidden=None, lengths=None, backward=False):
        max_time, batch_size, _ = input_.size()

        if hidden is None:
            h0 = input_.new(batch_size, self.hidden_size).zero_()
            c0 = input_.new(batch_size, self.hidden_size).zero_()
        else:
            h0, c0 = hidden
            if h0.dim() == 3:
                h0, c0 = h0.squeeze(0), c0.squeeze(0)

        if lengths is None:
            lengths = input_.new([max_time] * batch_size).long()

        output, (h_n, c_n) = CustomLSTM._forward_rnn(
            self.cell, input_, (h0, c0), lengths, backward=backward)

        return output, (h_n.unsqueeze(0), c_n.unsqueeze(0))


class CustomBiLSTM(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers=1, dropout=0.0):
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        super().__init__()

        layers = []
        rnn_inp = input_size
        for layer in range(num_layers):
            fwd = CustomLSTM(rnn_inp, hidden_size)
            bwd = CustomLSTM(rnn_inp, hidden_size)
            layers.append((fwd, bwd))
            self.add_module('fwd_{}'.format(layer), fwd)
            self.add_module('bwd_{}'.format(layer), bwd)
            rnn_inp = hidden_size * 2
        self.layers = layers

    def forward(self, inputs, hidden=None, lengths=None):
        if hidden is not None:
            h0, c0 = hidden
            h0, c0 = h0.view(len(self.layers), 2, -1), c0.view(len(self.layers), 2, -1)
            (fwd_h0, bwd_h0), (fwd_c0, bwd_c0) = h0.split(1, dim=1), c0.split(1, dim=1)
            fwd_hidden = tuple(zip(fwd_h0, fwd_c0))
            bwd_hidden = tuple(zip(fwd_h0, fwd_c0))
        else:
            fwd_hidden = bwd_hidden = [None] * len(self.layers)

        hn, cn = [], []
        for layer, (fwd, bwd) in enumerate(self.layers):
            fwd_outs, (fwd_h1, fwd_c1) = fwd(
                inputs, fwd_hidden[layer], lengths=lengths)
            bwd_outs, (bwd_h1, bwd_c1) = bwd(
                inputs, bwd_hidden[layer], lengths=lengths, backward=True)
            # compute new input (seq_len x batch x 2 * hidden_size)
            inputs = torch.cat([fwd_outs, bwd_outs], dim=2)
            # apply dropout
            if layer < self.num_layers - 1:
                inputs = torch_utils.sequential_dropout(
                    inputs, self.dropout, self.training)
            # store hidden
            hn.append(torch.cat([fwd_h1, bwd_h1], dim=0))
            cn.append(torch.cat([fwd_c1, bwd_c1], dim=0))

        return inputs, (torch.cat(hn, dim=0), torch.cat(cn, dim=0))




import torch


class Beam(object):
    """
    Beam class for performing beam search

    Parameters
    -----------
    width: int, beam buffer size. The higher the better the chances are of
        actually decoding the best sequence but also the bigger the memory
        footprint
    prev: int, integer token to use as first decoding step
    eos: int or None, integer corresponding to the <eos> symbol in the
        vocabulary. It will be used as terminating criterion for the decoding
    """
    def __init__(self, width, eos, bos=None, device='cpu'):
        # attributes
        self.width = width
        self.eos = eos
        self.bos = bos or eos

        # data
        self.active = True
        self.scores = torch.zeros(width)
        init_state = torch.zeros(width, dtype=torch.int64, device=device) + self.bos
        # output values at each beam
        self.beam_values = [init_state]
        # backpointer to previous beam
        self.source_beams = []

    def __len__(self):
        """
        number of steps already decoded
        """
        return len(self.source_beams)

    def _new_beam(self, outs):
        """
        Computes a new beam based on the current model output and the history.

        outs: (width x vocab)
        """
        if len(self) > 0:
            beam_outs = outs + self.scores.unsqueeze(1).expand_as(outs)
            # EOS nihilation (adapted from OpenNMT)
            for i in range(len(beam_outs)):
                if self.beam_values[-1][i] == self.eos:
                    beam_outs[i] = -1e20
        else:
            # all beams have same start values, just pick the 1st
            beam_outs = outs[0]

        width, vocab = outs.size()
        # compute best outputs over a flatten vector of size (width x vocab)
        # i.e. regardless their source beam
        scores, flatten_ids = beam_outs.view(-1).topk(self.width, dim=0)
        # compute source beam, best candidates (can't do // on tensors)
        source_beams, beam = flatten_ids / vocab, flatten_ids % vocab

        return scores, source_beams, beam

    def get_source_beam(self, step=-1):
        """
        Get entries in previous beam leading to a given step.
        """
        if step >= len(self.source_beams):
            raise ValueError("Only {} decoded steps".format(len(self.source_beams)))

        return self.source_beams[step]

    def get_current_state(self):
        """
        Get beam at current step
        """
        return self.beam_values[-1]

    def finished(self, beam):
        """
        Finished criterion based on whether the last best hypothesis is EOS
        """
        return beam[0] == self.eos

    def advance(self, outs):
        """
        Runs a decoder step accumulating the path and the ids.
        """
        scores, source_beams, beam = self._new_beam(outs)

        if self.finished(beam):
            self.active = False

        self.scores = scores
        self.source_beams.append(source_beams)
        self.beam_values.append(beam)

    def get_hypothesis(self, idx):
        """
        Get hypothesis for `idx` entry in the current beam step.
        Note that the beam isn't mantained in sorted order.
        """
        if idx > self.width:
            raise ValueError("Beam has only capacity {}".format(self.width))

        hypothesis = []
        for step in range(len(self) - 1, -1, -1):
            hypothesis.append(self.beam_values[step+1][idx].item())
            idx = self.get_source_beam(step=step)[idx]
        return hypothesis[::-1]

    def decode(self, n=1):
        """
        Get n best hypothesis at current step.
        """
        if n > self.width:
            raise ValueError("Beam has only capacity {}".format(self.width))

        scores, beam_ids = torch.sort(self.scores, dim=0, descending=True)
        best_scores, best_beam_ids = scores[:n].tolist(), beam_ids[:n].tolist()
        best_hyps = [self.get_hypothesis(b) for b in best_beam_ids]
        return best_scores, best_hyps




import difflib
import collections


Keep = collections.namedtuple('Keep', [])
Replace = collections.namedtuple('Replace', ['source', 'target'])
Insert = collections.namedtuple('Insert', ['string', 'pos'])
Delete = collections.namedtuple('Delete', ['string'])


def get_edit_scripts_(s1, s2):
    differ = difflib.Differ()

    # no need to compare
    if len(s1) == 1:
        yield Replace(source=s1, target=s2)
        return

    # diff edits
    acts = [(act, char) for (act, _, char) in differ.compare(s1, s2)]
    i = 0
    while i < len(acts):
        a, c = acts[i]

        if a == ' ':
            # if the are only additions left, we should merge here
            if acts[i+1:] and acts[i+1][0] == '+':
                all_adds, cs = True, ''
                for (a, cc) in acts[i+1:]:
                    if a == '+':
                        cs += cc
                    elif a == '-':
                        all_adds = False
                        break
                    else:
                        break
                if all_adds:
                    yield Insert(string=cs, pos='post')
                    i += len(cs) + 1
                else:
                    yield Keep()
                    i += 1
            else:
                yield Keep()
                i += 1

        else:
            # grab edits involving any number of -/+ followed by any number of the other edit operation
            merge_a, merge_b = [(a, c)], []
            flipped = False
            for j in range(i + 1, len(acts)):
                a2, c2 = acts[j]
                if a2 == ' ':
                    break
                elif a2 == a:
                    if flipped:
                        break
                    else:
                        merge_a.append((a2, c2))
                else:
                    flipped = True
                    merge_b.append((a2, c2))

            # put always the substract first
            if merge_a[0][0] == '-':
                deletes, adds = merge_a, merge_b
            else:
                deletes, adds = merge_b, merge_a
            del merge_a, merge_b

            # do the actual merging
            if len(deletes) > len(adds):
                # zip and delete the remaining items
                for (_, c_b), (_, c_a) in zip(deletes, adds):
                    yield Replace(source=c_b, target=c_a)
                for (a, c) in deletes[len(adds):]:
                    yield Delete(string=c)

            elif len(adds) > len(deletes):
                # zip as many as possible and merge the remaining additions
                if deletes:
                    for idx in range(len(deletes) - 1):
                        (_, c_a), (_, c_b) = adds[idx], deletes[idx]
                        yield Replace(source=c_b, target=c_a)
                    _, c_b = deletes[-1]
                    c_a = ''.join([c for _, c in adds[len(deletes)-1:]])
                    yield Replace(source=c_b, target=c_a)
                else:
                    # this must mean that the additions are followed by a keep
                    if acts[i+len(adds)][0] == ' ':
                        c_a = ''.join([c for _, c in adds])
                        yield Insert(string=c_a, pos="pre")
                        i += 1
                    else:
                        raise ValueError("Expected a bunch of additions followed by keep")
            else:
                # same size, just zip them into replacements
                for (_, c_b), (_, c_a) in zip(deletes, adds):
                    yield Replace(source=c_b, target=c_a)

            i += len(adds) + len(deletes)


def transform(lem, tok):
    return tuple(get_edit_scripts_(tok, lem))


def inverse_transform(pred, tok):
    output, idx = "", 0
    for rule in pred:
        if isinstance(rule, Keep):
            output += tok[idx]
            idx += 1
        elif isinstance(rule, Replace):
            output += rule.target
            idx += len(rule.source)
        elif isinstance(rule, Insert):
            if rule.pos == 'pre':
                output += rule.string + tok[idx]
            else:
                output += tok[idx] + rule.string
            idx += 1
        else:
            idx += len(rule.string)
    return output




import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_packed_sequence as unpack

from pie import initialization
from pie import torch_utils
from pie.constants import TINY

from .beam_search import Beam
from .attention import Attention
from .highway import Highway


class LinearDecoder(nn.Module):
    """
    Simple Linear Decoder that outputs a probability distribution
    over the vocabulary

    Parameters
    ===========
    label_encoder : LabelEncoder
    in_features : int, input dimension
    """
    def __init__(self, label_encoder, in_features, dropout=0.0,
                 highway_layers=0, highway_act='relu'):
        self.label_encoder = label_encoder
        super().__init__()

        # nll weight
        nll_weight = torch.ones(len(label_encoder))
        nll_weight[label_encoder.get_pad()] = 0.
        self.register_buffer('nll_weight', nll_weight)
        # highway
        self.highway = None
        if highway_layers > 0:
            self.highway = Highway(in_features, highway_layers, highway_act)
        # decoder output
        self.decoder = nn.Linear(in_features, len(label_encoder))
        self.init()

    def init(self):
        # linear
        initialization.init_linear(self.decoder)

    def forward(self, enc_outs):
        if self.highway is not None:
            enc_outs = self.highway(enc_outs)
        linear_out = self.decoder(enc_outs)

        return linear_out

    def loss(self, logits, targets):
        loss = F.cross_entropy(
            logits.view(-1, len(self.label_encoder)), targets.view(-1),
            weight=self.nll_weight, reduction="mean",
            ignore_index=self.label_encoder.get_pad())

        return loss

    def predict(self, enc_outs, lengths):
        """
        Parameters
        ==========
        enc_outs : torch.tensor(seq_len x batch x hidden_size)
        """
        if self.highway is not None:
            enc_outs = self.highway(enc_outs)
        probs = F.softmax(self.decoder(enc_outs), dim=-1)
        probs, preds = torch.max(probs.transpose(0, 1), dim=-1)
        output_probs, output_preds = [], []
        for idx, length in enumerate(lengths.tolist()):
            output_preds.append(
                self.label_encoder.inverse_transform(preds[idx])[:length])
            output_probs.append(probs[idx].tolist())

        return output_preds, output_probs


class CRFDecoder(nn.Module):
    """
    CRF decoder layer
    """
    def __init__(self, label_encoder, hidden_size, highway_layers=0, highway_act='relu'):
        self.label_encoder = label_encoder
        super().__init__()

        vocab = len(label_encoder)
        self.highway = None
        if highway_layers > 0:
            self.highway = Highway(hidden_size, highway_layers, highway_act)
        self.projection = nn.Linear(hidden_size, vocab)
        self.transition = nn.Parameter(torch.Tensor(vocab, vocab))
        self.start_transition = nn.Parameter(torch.Tensor(vocab))
        self.end_transition = nn.Parameter(torch.Tensor(vocab))

        self.init()

    def init(self):
        # transitions
        nn.init.xavier_normal_(self.transition)
        nn.init.normal_(self.start_transition)
        nn.init.normal_(self.end_transition)

    def forward(self, enc_outs):
        """get logits of the input features"""
        # (seq_len x batch x vocab)
        if self.highway is not None:
            enc_outs = self.highway(enc_outs)
        logits = self.projection(enc_outs)

        return F.log_softmax(logits, -1)

    def partition(self, logits, mask):
        seq_len, batch, vocab = logits.size()

        # (batch x vocab)
        Z = self.start_transition.view(1, vocab) + logits[0]

        for t in range(1, seq_len):
            emit_score = logits[t].view(batch, 1, vocab)
            trans_score = self.transition.view(1, vocab, vocab)
            # mask and update
            Z_t = Z.view(batch, vocab, 1)
            # (batch x vocab x vocab) => (batch x vocab)
            Z_t = torch_utils.log_sum_exp(Z_t + emit_score + trans_score, 1)
            # mask & update
            mask_t = mask[t].view(batch, 1)
            Z = Z_t * mask_t + Z * (1 - mask_t)

        Z = Z + self.end_transition.view(1, vocab)
        # (batch x vocab) => (batch)
        Z = torch_utils.log_sum_exp(Z)

        return Z

    def score(self, logits, mask, targets):
        # calculate the score of a given sequence
        seq_len, batch, vocab = logits.size()
        # transition from start tag to first tag
        score = self.start_transition.index_select(0, targets[0])
        # (batch x vocab x vocab)
        trans = self.transition.unsqueeze(0).expand(batch, vocab, vocab)

        # iterate from tag to next tag
        for t in range(seq_len - 1):
            curr_tag, next_tag = targets[t], targets[t+1]
            # from current transition scores (batch, 1, vocab) => (batch, vocab)
            trans_score = trans.gather(
                1, curr_tag.view(batch, 1, 1).expand(batch, 1, vocab)
            ).squeeze(1)
            # from current to next transition scores (batch, 1) => (batch)
            trans_score = trans_score.gather(1, next_tag.view(batch, 1)).squeeze(1)
            # (batch)
            emit_score = logits[t].gather(1, curr_tag.view(batch, 1)).squeeze(1)
            score = score + (trans_score * mask[t+1]) + (emit_score * mask[t])
            
        # last step
        last_target_index = mask.sum(0).long() - 1
        # (batch)
        last_target = targets.gather(0, last_target_index.expand(seq_len, batch))[0]
        # (batch)
        last_trans_score = self.end_transition.index_select(0, last_target)
        # (batch x 1) => (batch)
        last_emit_score = logits[-1].gather(1, last_target.unsqueeze(1)).squeeze(1)
        # (batch)
        score = score + last_trans_score + last_emit_score * mask[-1]

        return score

    def loss(self, logits, targets, lengths):
        # mask on padding
        mask = torch_utils.make_length_mask(lengths).float()
        # (batch x seq_len) => (seq_len x batch)
        mask = mask.t()
        # logits = logits * mask.unsqueeze(2).expand_as(logits)
        Z = self.partition(logits, mask)
        score = self.score(logits, mask, targets)

        # FIXME: this gives the average loss per sentence (perhaps it should)
        # be weighted down to make it also per word?
        return torch.mean(Z - score)

    def predict(self, enc_outs, lengths):
        # (seq_len x batch x vocab)
        logits = self(enc_outs)
        seq_len, _, vocab = logits.size()
        start_tag, end_tag = vocab, vocab + 1

        # mask on padding (batch x seq_len)
        mask = torch_utils.make_length_mask(lengths).float()
        # Mask is not used !

        # variables
        trans = logits.new(vocab + 2, vocab + 2).fill_(-10000.)
        trans[:vocab, :vocab] = self.transition.data
        hyps, scores = [], []
        tag_sequence = logits.new(seq_len + 2, vocab + 2)

        # iterate over batches
        for logits_b, len_b in zip(logits.transpose(0, 1), lengths):
            seq_len = len_b.item()
            # get this batch logits
            tag_sequence.fill_(-10000)
            tag_sequence[0, start_tag] = 0.
            tag_sequence[1:seq_len+1, :vocab] = logits_b[:seq_len]
            tag_sequence[seq_len+1, end_tag] = 0.

            path, score = torch_utils.viterbi_decode(tag_sequence[:seq_len+2], trans)
            hyps.append(self.label_encoder.inverse_transform(path[1:-1]))
            scores.append(score)

        return hyps, scores


class AttentionalDecoder(nn.Module):
    """
    Decoder using attention over the entire input sequence

    Parameters
    ===========
    label_encoder : LabelEncoder of the task
    in_dim : int, embedding dimension of the task.
        It should be the same as the corresponding encoder to ensure that weights
        can be shared.
    hidden_size : int, hidden size of the encoder, decoder and attention modules.
    context_dim : int (optional), dimensionality of additional context vectors
    """
    def __init__(self, label_encoder, in_dim, hidden_size, dropout=0.0,
                 # rnn
                 num_layers=1, cell='LSTM', init_rnn='default',
                 # attention
                 scorer='general',
                 # sentence context
                 context_dim=0):

        self.label_encoder = label_encoder
        self.context_dim = context_dim
        self.num_layers = num_layers
        self.dropout = dropout
        self.init_rnn = init_rnn
        super().__init__()

        if label_encoder.get_eos() is None or label_encoder.get_bos() is None:
            raise ValueError("AttentionalDecoder needs <eos> and <bos>")

        # nll weight
        nll_weight = torch.ones(len(label_encoder))
        nll_weight[label_encoder.get_pad()] = 0.
        self.register_buffer('nll_weight', nll_weight)
        # emb
        self.embs = nn.Embedding(len(label_encoder), in_dim)
        # rnn
        self.rnn = getattr(nn, cell)(in_dim + context_dim, hidden_size,
                                     num_layers=num_layers,
                                     dropout=dropout if num_layers > 1 else 0)
        self.attn = Attention(hidden_size)
        self.proj = nn.Linear(hidden_size, len(label_encoder))

        self.init()

    def init(self):
        # embeddings
        initialization.init_embeddings(self.embs)
        # rnn
        initialization.init_rnn(self.rnn, scheme=self.init_rnn)
        # linear
        initialization.init_linear(self.proj)

    def forward(self, targets, lengths, enc_outs, src_lengths, context=None):
        """
        Decoding routine for training. Returns the logits corresponding to
        the targets for the `loss` method. Takes care of padding.
        """
        targets, lengths = targets[:-1], lengths - 1
        embs = self.embs(targets)

        if self.context_dim > 0:
            if context is None:
                raise ValueError("Contextual Decoder needs `context`")
            # (seq_len x batch x emb_dim) + (batch x context_dim)
            embs = torch.cat(
                [embs, context.unsqueeze(0).repeat(embs.size(0), 1, 1)],
                dim=2)

        embs, unsort = torch_utils.pack_sort(embs, lengths)

        outs, _ = self.rnn(embs)
        outs, _ = unpack(outs)
        outs = outs[:, unsort]

        context, _ = self.attn(outs, enc_outs, src_lengths)

        return self.proj(context)

    def loss(self, logits, targets):
        """
        Compute loss from logits (output of forward)

        Parameters
        ===========
        logits : tensor(seq_len x batch x vocab)
        targets : tensor(seq_len x batch)
        """
        targets = targets[1:]  # remove <bos> from targets

        loss = F.cross_entropy(
            logits.view(-1, len(self.label_encoder)), targets.view(-1),
            weight=self.nll_weight, reduction="mean",
            ignore_index=self.label_encoder.get_pad())

        # FIXME: normalize loss to be word-level

        return loss

    def predict_max(self, enc_outs, lengths,
                    max_seq_len=20, bos=None, eos=None,
                    context=None):
        """
        Decoding routine for inference with step-wise argmax procedure

        Parameters
        ===========
        enc_outs : tensor(src_seq_len x batch x hidden_size)
        context : tensor(batch x hidden_size), optional
        """
        eos = eos or self.label_encoder.get_eos()
        bos = bos or self.label_encoder.get_bos()
        hidden, batch, device = None, enc_outs.size(1), enc_outs.device
        mask = torch.ones(batch, dtype=torch.int64, device=device)
        inp = torch.zeros(batch, dtype=torch.int64, device=device) + bos
        hyps, scores = [], 0

        for _ in range(max_seq_len):
            if mask.sum().item() == 0:
                break

            # prepare input
            emb = self.embs(inp)
            if context is not None:
                emb = torch.cat([emb, context], dim=1)
            # run rnn
            emb = emb.unsqueeze(0)
            outs, hidden = self.rnn(emb, hidden)
            outs, _ = self.attn(outs, enc_outs, lengths)
            outs = self.proj(outs).squeeze(0)
            # get logits
            probs = F.log_softmax(outs, dim=1)
            # sample and accumulate
            score, inp = probs.max(1)
            hyps.append(inp.tolist())
            mask = mask * (inp != eos).long()
            score = score.cpu()
            score[mask == 0] = 0
            scores += score

        hyps = [self.label_encoder.stringify(hyp) for hyp in zip(*hyps)]
        scores = [s/(len(hyp) + TINY) for s, hyp in zip(scores.tolist(), hyps)]

        return hyps, scores

    def predict_beam(self, enc_outs, lengths,
                     max_seq_len=50, width=12, eos=None, bos=None,
                     context=None):
        """
        Decoding routine for inference with beam search

        Parameters
        ===========
        enc_outs : tensor(src_seq_len x batch x hidden_size)
        context : tensor(batch x hidden_size), optional
        """
        eos = eos or self.label_encoder.get_eos()
        bos = bos or self.label_encoder.get_bos()
        hidden, device = None, enc_outs.device
        seq_len, batch, _ = enc_outs.size()
        beams = [Beam(width, eos=eos, bos=bos, device=device) for _ in range(batch)]

        # expand data along beam width
        # (seq_len x beam * batch x hidden_size)
        enc_outs = enc_outs.repeat(1, width, 1)
        lengths = lengths.repeat(width)
        if context is not None:
            # (beam * batch x context_dim)
            context = context.repeat(width, 1)

        for _ in range(max_seq_len):
            if all(not beam.active for beam in beams):
                break
            # (beam x batch)
            inp = torch.stack([beam.get_current_state() for beam in beams], dim=1)
            # (beam * batch)
            inp = inp.view(-1)
            # (beam * batch x emb_dim)
            emb = self.embs(inp)
            if context is not None:
                # (beam * batch x emb_dim + context_dim)
                emb = torch.cat([emb, context], dim=1)
            # run rnn
            emb = emb.unsqueeze(0)  # add singleton seq_len dim
            outs, hidden = self.rnn(emb, hidden)

            # (1 x beam * batch x hidden)
            outs, _ = self.attn(outs, enc_outs, lengths)
            # (beam * batch x vocab)
            outs = self.proj(outs).squeeze(0)
            # (beam * batch x vocab)
            probs = F.log_softmax(outs, dim=1)
            # (beam x batch x vocab)
            probs = probs.view(width, batch, -1)

            # expose beam dim for swaping
            if isinstance(hidden, tuple):
                hidden = hidden[0].view(self.num_layers, width, batch, -1), \
                         hidden[1].view(self.num_layers, width, batch, -1)
            else:
                hidden = hidden.view(self.num_layers, width, batch, -1)

            # advance and swap
            for i, beam in enumerate(beams):
                if not beam.active:
                    continue
                # advance
                beam.advance(probs[:, i])
                # rearrange
                sbeam = beam.get_source_beam()
                if isinstance(hidden, tuple):
                    hidden[0][:, :, i].copy_(hidden[0][:, :, i].index_select(1, sbeam))
                    hidden[1][:, :, i].copy_(hidden[1][:, :, i].index_select(1, sbeam))
                else:
                    hidden[:, :, i].copy_(hidden[:, :, i].index_select(1, sbeam))

            # collapse beam and batch
            if isinstance(hidden, tuple):
                hidden = hidden[0].view(self.num_layers, width * batch, -1), \
                         hidden[1].view(self.num_layers, width * batch, -1)
            else:
                hidden = hidden.view(self.num_layers, width * batch, -1)

        scores, hyps = [], []
        for beam in beams:
            bscores, bhyps = beam.decode(n=1)
            bscores, bhyps = bscores[0], bhyps[0]
            scores.append(bscores)
            hyps.append(bhyps)

        hyps = [self.label_encoder.stringify(hyp) for hyp in hyps]
        scores = [s/(len(hyp) + TINY) for s, hyp in zip(scores, hyps)]

        return hyps, scores




import os
import json
import tarfile
import logging

import tqdm
import torch
import torch.nn as nn

from pie import utils
from pie.data import MultiLabelEncoder
from pie.settings import Settings

from .scorer import Scorer, get_known_and_ambigous_tokens


class BaseModel(nn.Module):
    """
    Abstract model class defining the model interface
    """
    def __init__(self, label_encoder, tasks, *args, **kwargs):
        self.label_encoder = label_encoder
        # prepare input task data from task settings
        if isinstance(tasks, list):
            tasks = {task['name']: task for task in tasks}
        self.tasks = tasks
        self.known = set()
        self.ambs = {task: set() for task in tasks}
        self._fitted_trainset_scorer = False
        super().__init__()

    def get_scorer(self, task, trainset=None):
        """ Given a task, gets a scorer. Trainset can be used for computing
        unknown and ambiguous tokens.

        :param task: Taskname (str)
        :param trainset: Dataset for training
        :return: Scorer
        """
        scorer = Scorer(self.label_encoder.tasks[task])
        if not self._fitted_trainset_scorer and trainset:
            self.known, self.ambs = get_known_and_ambigous_tokens(
                trainset, list(self.label_encoder.tasks.values()))
            self._fitted_trainset_scorer = True
        scorer.set_known_and_amb(self.known, self.ambs[task])
        return scorer

    def loss(self, batch_data):
        """
        """
        raise NotImplementedError

    def predict(self, inp, *tasks, **kwargs):
        """
        Compute predictions based on already processed input
        """
        raise NotImplementedError

    def get_args_and_kwargs(self):
        """
        Return a dictionary of {'args': tuple, 'kwargs': dict} that were used
        to instantiate the model (excluding the label_encoder and tasks)
        """
        raise NotImplementedError

    def evaluate(self, dataset, trainset=None, **kwargs):
        """
        Get scores per task

        dataset: pie.data.Dataset, dataset to evaluate on (your dev or test set)
        trainset: pie.data.Dataset (optional), if passed scores for unknown and ambiguous
            tokens can be computed
        **kwargs: any other arguments to Model.predict
        """
        assert not self.training, "Ooops! Inference in training mode. Call model.eval()"

        scorers = {task: self.get_scorer(task, trainset) for task in self.tasks}

        with torch.no_grad():
            for (inp, tasks), (rinp, rtasks) in tqdm.tqdm(
                    dataset.batch_generator(return_raw=True)):

                preds = self.predict(inp, **kwargs)

                # - get input tokens
                tokens = [w for line in rinp for w in line]

                # - get trues
                trues = {}
                for task in preds:
                    le = self.label_encoder.tasks[task]
                    # - transform targets
                    trues[task] = le.preprocess(
                        [t for line in rtasks for t in line[le.target]], tokens)

                    # - flatten token level predictions
                    if le.level == 'token':
                        preds[task] = [pred for batch in preds[task] for pred in batch]

                # accumulate
                for task, scorer in scorers.items():
                    scorer.register_batch(preds[task], trues[task], tokens)

        return scorers

    def save(self, fpath, infix=None, settings=None):
        """
        Serialize model to path
        """
        import pie
        fpath = utils.ensure_ext(fpath, 'tar', infix)

        # create dir if necessary
        dirname = os.path.dirname(fpath)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        with tarfile.open(fpath, 'w') as tar:
            # serialize label_encoder
            string = json.dumps(self.label_encoder.jsonify())
            path = 'label_encoder.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize tasks
            string, path = json.dumps(self.tasks), 'tasks.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize model class
            string, path = str(type(self).__name__), 'class.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize parameters
            string, path = json.dumps(self.get_args_and_kwargs()), 'parameters.zip'
            utils.add_gzip_to_tar(string, path, tar)

            # serialize weights
            utils.add_weights_to_tar(self.state_dict(), 'state_dict.pt', tar)

            # serialize current pie commit
            if pie.__commit__ is not None:
                string, path = pie.__commit__, 'pie-commit.zip'
                utils.add_gzip_to_tar(string, path, tar)

            # if passed, serialize settings
            if settings is not None:
                string, path = json.dumps(settings), 'settings.zip'
                utils.add_gzip_to_tar(string, path, tar)

        return fpath

    @staticmethod
    def load_settings(fpath):
        """
        Load settings from path
        """
        with tarfile.open(utils.ensure_ext(fpath, 'tar'), 'r') as tar:
            return Settings(json.loads(utils.get_gzip_from_tar(tar, 'settings.zip')))

    @staticmethod
    def load(fpath):
        """
        Load model from path
        """
        import pie

        with tarfile.open(utils.ensure_ext(fpath, 'tar'), 'r') as tar:
            # check commit
            try:
                commit = utils.get_gzip_from_tar(tar, 'pie-commit.zip')
            except Exception:
                commit = None
            if (pie.__commit__ and commit) and pie.__commit__ != commit:
                logging.warn(
                    ("Model {} was serialized with a previous "
                     "version of `pie`. This might result in issues. "
                     "Model commit is {}, whereas current `pie` commit is {}.").format(
                         fpath, commit, pie.__commit__))

            # load label encoder
            le = MultiLabelEncoder.load_from_string(
                utils.get_gzip_from_tar(tar, 'label_encoder.zip'))

            # load tasks
            tasks = json.loads(utils.get_gzip_from_tar(tar, 'tasks.zip'))

            # load model parameters
            params = json.loads(utils.get_gzip_from_tar(tar, 'parameters.zip'))

            # instantiate model
            model_type = getattr(pie.models, utils.get_gzip_from_tar(tar, 'class.zip'))
            with utils.shutup():
                model = model_type(le, tasks, *params['args'], **params['kwargs'])

            # load settings
            try:
                settings = Settings(
                    json.loads(utils.get_gzip_from_tar(tar, 'settings.zip')))
                model._settings = settings
            except Exception:
                logging.warn("Couldn't load settings for model {}!".format(fpath))

            # load state_dict
            model.load_state_dict(
                torch.load(tar.extractfile('state_dict.pt'), map_location='cpu'))

        model.eval()

        return model




import yaml
import difflib
from termcolor import colored
from terminaltables import github_table
from collections import Counter, defaultdict

from sklearn.metrics import accuracy_score, precision_recall_fscore_support
from sklearn.utils.multiclass import unique_labels
import numpy as np
from pie import utils
from pie import constants


def get_known_and_ambigous_tokens(trainset, label_encoders):
    """ Retrieve known and ambiguous token for all label encoders, for one dataset

    :param trainset: Trainset
    :param label_encoders: List of label encoders
    :return: Known set of tokens, ambiguous token par task (Dict[task, Set[str])
    """
    known = set()
    ambs = defaultdict(lambda: defaultdict(Counter))
    targets = [le.target for le in label_encoders]
    for _, (inp, tasks) in trainset.reader.readsents():
        known.update(inp)
        for le in label_encoders:
            for tok, true in zip(inp, le.preprocess(tasks[le.target], inp)):
                ambs[le.target][tok][true] += 1
    ambs = {t: set(tok for tok in ambs[t] if len(ambs[t][tok]) > 1) for t in ambs}
    return known, ambs


def compute_scores(trues, preds):

    def format_score(score):
        return round(float(score), 4)

    with utils.shutup():
        p, r, f1, _ = precision_recall_fscore_support(trues, preds, average="macro")
        p = format_score(p)
        r = format_score(r)
        a = format_score(accuracy_score(trues, preds))

    return {'accuracy': a, 'precision': p, 'recall': r, 'support': len(trues)}


class Scorer(object):
    """
    Accumulate predictions over batches and compute evaluation scores
    """
    def __init__(self, label_encoder):
        self.label_encoder = label_encoder
        self.known_tokens = self.amb_tokens = None
        self.preds = []
        self.trues = []
        self.tokens = []

    def set_known_and_amb(self, known_tokens, amb_tokens):
        """ Set known tokens as well as ambiguous tokens """
        self.known_tokens = known_tokens
        self.amb_tokens = amb_tokens

    def serialize_preds(self, path):
        """
        write predictions to file
        """
        with open(path, 'w') as f:
            for tok, true, pred in zip(self.tokens, self.trues, self.preds):
                f.write('{}\t{}\t{}\n'.format(tok, true, pred))

    def register_batch(self, hyps, targets, tokens):
        """
        hyps : list
        targets : list
        tokens : list
        """
        if len(hyps) != len(targets) or len(targets) != len(tokens):
            raise ValueError("Unequal input lengths. Hyps {}, targets {}, tokens {}"
                             .format(len(hyps), len(targets), len(tokens)))

        for pred, true, token in zip(hyps, targets, tokens):
            if self.label_encoder.preprocessor_fn:
                true = self.label_encoder.preprocessor_fn.inverse_transform(true, token)
                try:
                    pred = self.label_encoder.preprocessor_fn.inverse_transform(
                        pred, token)
                except:
                    pred = constants.INVALID

            self.preds.append(pred)
            self.trues.append(true)
            self.tokens.append(token)

    def get_scores(self):
        """
        Return a dictionary of scores
        """
        output = {}
        output['all'] = compute_scores(self.trues, self.preds)

        # apply text transformations to known tokens
        known_targets = None
        if self.label_encoder.known_tokens:
            known_targets = set(self.label_encoder.preprocess_text(
                list(self.label_encoder.known_tokens)))

        # compute scores for unknown input tokens
        unk_trues, unk_preds, amb_trues, amb_preds = [], [], [], []
        knw_trues, knw_preds, unk_trg_trues, unk_trg_preds = [], [], [], []
        for true, pred, token in zip(self.trues, self.preds, self.tokens):
            if self.known_tokens and token in self.known_tokens:
                knw_trues.append(true)
                knw_preds.append(pred)
            if self.known_tokens and token not in self.known_tokens:
                unk_trues.append(true)
                unk_preds.append(pred)
            if self.amb_tokens and token in self.amb_tokens:
                amb_trues.append(true)
                amb_preds.append(pred)
            # token-level encoding doesn't have unknown targets (only OOV)
            if known_targets and true not in known_targets:
                unk_trg_trues.append(true)
                unk_trg_preds.append(pred)

        output['known-tokens'] = compute_scores(knw_trues, knw_preds)
        support = len(unk_trues)
        if support > 0:
            output['unknown-tokens'] = compute_scores(unk_trues, unk_preds)
        support = len(amb_trues)
        if support > 0:
            output['ambiguous-tokens'] = compute_scores(amb_trues, amb_preds)
        support = len(unk_trg_trues)
        if support > 0:
            output['unknown-targets'] = compute_scores(unk_trg_trues, unk_trg_preds)

        return output

    def get_most_common(self, errors, most_common):
        # sort by number of target errors
        def key(item): target, errors = item; return sum(errors.values())

        errors = dict(sorted(errors.items(), key=key, reverse=True))

        output = []
        for true, preds in errors.items():
            if len(output) >= most_common:
                break
            output.append((true, sum(preds.values()), preds))

        return output

    def get_confusion_matrix(self):
        errors = defaultdict(Counter)
        for true, pred in zip(self.trues, self.preds):
            if true != pred:
                errors[true][pred] += 1

        return errors

    def get_confusion_matrix_table(self) -> list:
        """
        Returns a table formated confusion matrix
        """
        matrix = self.get_confusion_matrix()
        table = []
        # Retrieve each true prediction and its dictionary of errors
        for expected, pred_counter in matrix.items():
            counts = [(word, counter) for word, counter in sorted(
                pred_counter.items(), key=lambda tup: tup[1], reverse=True)]
            total = sum(pred_counter.values())
            table.append((expected, total, counts))
        # Sort by error sum
        table = sorted(table, reverse=True, key=lambda tup: tup[1])
        # Then, we expand lines
        output = []
        for word, total, errors in table:
            for index, (prediction, counter) in enumerate(errors):
                row = ["", ""]
                if index == 0:
                    row = [word, total]
                row += [prediction, counter]
                output.append(row)
        return [["Expected", "Total Errors", "Predictions", "Predicted times"]] + output

    def get_classification_summary(self, most_common=200):
        """
        Get a printable summary for classification errors
        """
        errors = self.get_confusion_matrix()

        output = ''
        for true, counts, preds in self.get_most_common(errors, most_common):
            true = '{}(#{})'.format(colored(true, 'green'), counts)
            preds = Counter(preds).most_common(10)
            preds = '\t'.join('{}'.format('{}(#{})'.format(p, c)) for p, c in preds)
            output += '{}\t{}\n'.format(true, preds)

        return output

    def get_transduction_summary(self, most_common=100):
        """
        Get a printable summary of string transduction errors
        """
        COLORS = {' ': 'white', '+': 'green', '-': 'red'}

        def get_diff(true, pred):
            diff = ''
            for action, *_, char in difflib.Differ().compare(true, pred):
                diff += colored(char, COLORS[action])
            return diff

        def error_summary(true, count, preds):
            summary = '{}(#{})\n'.format(true, count)
            summary += ''.join(
                '\t{:<20}\t{}\n'.format(
                    '{}(#{})'.format(pred, pcount), get_diff(true, pred))
                for pred, pcount in preds.items())
            return summary

        known_targets = self.label_encoder.known_tokens
        known, unk_trg = defaultdict(Counter), defaultdict(Counter)
        unk_tok, amb_tok = defaultdict(Counter), defaultdict(Counter)
        for true, pred, token in zip(self.trues, self.preds, self.tokens):
            if true == pred:
                continue
            if self.known_tokens and token in self.known_tokens:
                known[true][pred] += 1
            if known_targets and true not in known_targets:
                unk_trg[true][pred] += 1
            if self.known_tokens and token not in self.known_tokens:
                unk_tok[true][pred] += 1
            if self.amb_tokens and token in self.amb_tokens:
                amb_tok[true][pred] += 1

        summary = []
        summary_ = '::: Known tokens :::\n\n'
        for true, count, preds in self.get_most_common(known, most_common):
            summary_ += '{}\n'.format(error_summary(true, count, preds))
        summary += [summary_]
        if unk_trg:
            summary_ = '::: Unknown targets :::\n\n'
            for true, count, preds in self.get_most_common(unk_trg, most_common):
                summary_ += '{}\n'.format(error_summary(true, count, preds))
            summary += [summary_]
        if amb_tok:
            summary_ = '::: Ambiguous tokens :::\n\n'
            for true, count, preds in self.get_most_common(amb_tok, most_common):
                summary_ += '{}\n'.format(error_summary(true, count, preds))
            summary += [summary_]
        if unk_tok:
            summary_ = '::: Unknown tokens :::\n\n'
            for true, count, preds in self.get_most_common(unk_tok, most_common):
                summary_ += '{}\n'.format(error_summary(true, count, preds))
            summary += [summary_]

        return '\n'.join(summary)

    def print_summary(self, full=False, most_common=100, confusion_matrix=False,
                      scores=None, report=False, markdown=True):
        """
        Get evaluation summary

        :param full: Get full report with error summary
        :param confusion_matrix: Get a confusion matrix
        :param most_common: Limit the full report to the number indicated
        :param scores: If scores are already computed, get passed here
        """

        print()
        if markdown:
            print("## " + self.label_encoder.name)
        else:
            print("::: Evaluation report for task: {} :::".format(
                self.label_encoder.name))
        print()

        if scores is None:
            scores = self.get_scores()

        # print scores
        if markdown:
            print(self.scores_in_markdown(scores) + '\n')
        else:
            print(yaml.dump(scores, default_flow_style=False))

        if full:
            print()
            if markdown:
                print("### Error summary for task {}".format(self.label_encoder.name))
            else:
                print("::: Error summary for task: {} :::".format(
                    self.label_encoder.name))
            print()
            if self.label_encoder.level == 'char':
                print(self.get_transduction_summary(most_common=most_common))
            else:
                print(self.get_classification_summary(most_common=most_common))

        if report:
            print()
            if markdown:
                print("### {} Classification report".format(self.label_encoder.name))
            else:
                print("::: Classification report :::")
            print()
            print(self.get_classification_report())

        if confusion_matrix:
            print()
            if markdown:
                print("### {} Confusion Matrix".format(self.label_encoder.name))
            else:
                print("::: Confusion Matrix :::")
            print()
            print(github_table.GithubFlavoredMarkdownTable(
                self.get_confusion_matrix_table()).table)

    def get_classification_report(self):
        return classification_report(
            y_true=self.trues,
            y_pred=self.preds)

    @staticmethod
    def scores_in_markdown(scores):
        measures = ["accuracy", "precision", "recall", "support"]
        table = [[""] + measures]
        for key in scores:
            table.append([key, *[scores[key][meas] for meas in measures]])

        return (github_table.GithubFlavoredMarkdownTable(table)).table


def classification_report(y_true, y_pred, digits=2):
    """ Generate a classification report similar to
    sklearn.metrics.classification_report but in markdown

    :param y_true: List of GT values
    :param y_pred: List of predictions
    :param digits: Number of float digits
    :return: Github Markdown Table
    """
    floatfmt = "{0:" + '.{:}f'.format(digits) + "}"

    labels = unique_labels(y_true, y_pred)
    target_names = [str(key) for key in labels]

    last_line_heading = 'avg / total'
    headers = ["target", "precision", "recall", "f1-score", "support"]

    p, r, f1, s = precision_recall_fscore_support(y_true, y_pred, average=None)

    formatted = []
    for nb_list in [p, r, f1]:
        formatted.append([floatfmt.format(x) for x in nb_list.tolist()])
    support = [[str(x) for x in s.tolist()]]

    tbl_rows = list(zip(target_names, *formatted, *support))

    # compute averages
    last_row = [last_line_heading,
                floatfmt.format(np.average(p)),
                floatfmt.format(np.average(r)),
                floatfmt.format(np.average(f1)),
                str(np.sum(s))]
    tbl_rows.append(last_row)

    return github_table.GithubFlavoredMarkdownTable([headers] + tbl_rows).table




import pie
import yaml

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('modelpath')
    args = parser.parse_args()

    m = pie.SimpleModel.load(args.modelpath)
    print("::: Settings :::")
    print(yaml.dump(dict(m._settings)))
    print()
    print("::: Model :::")
    print(m)





import os
import yaml
import json
from json_minify import json_minify

from pie import utils


DEFAULTPATH = os.sep.join([os.path.dirname(__file__), 'default_settings.json'])


class Settings(dict):
    def __init__(self, *args, **kwargs):
        super(Settings, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.iteritems():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Settings, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Settings, self).__delitem__(key)
        del self.__dict__[key]


def merge_task_defaults(settings):
    for task in settings.tasks:
        task_settings = task.get("settings", {})
        task_settings["target"] = task_settings.get("target", task['name'])
        task['settings'] = task_settings
        for tkey, tval in settings.task_defaults.items():
            task[tkey] = task.get(tkey, tval)

    return settings


def load_default_settings():
    """
    Load built-in default settings
    """
    with open(DEFAULTPATH) as f:
        return merge_task_defaults(Settings(json.loads(json_minify(f.read()))))


def parse_key(keys, v, defaults):
    """
    >>> parse_key(['a', 'b'], '1', {'a': {'b': 2}})
    {'a': {'b': 1}}
    """
    key, *keys = keys
    if key not in defaults:
        raise ValueError("Unknown key: ", key)
    if not keys:
        return {key: type(defaults[key])(v)}
    return {key: parse_key(keys, v, defaults[key])}


def parse_env_settings(defaults):
    output = {}
    for k, v in os.environ.items():
        if not k.startswith('PIE_'):
            continue
        keys = k.lower()[4:].split('__')
        output = utils.recursive_merge(output, parse_key(keys, v, defaults))

    return output


def check_settings(settings):
    has_target = False
    tasks = set(task['name'] for task in settings.tasks)

    for task in settings.tasks:
        # - check input char embeddings for attentional decoder
        if task['decoder'] == 'attentional':
            if settings.cemb_type.lower() not in ('rnn', 'cnn'):
                raise ValueError("Attentional decoder needs character embeddings")
        # - check conditions
        for task2 in task.get('conditions', []):
            if task2 not in tasks:
                raise ValueError("Task '{}' requires task '{}'".format(task, task2))
        # - check at least and at most one target
        if len(settings.tasks) == 1:
            task['target'] = True
        if task.get('target', False):
            if has_target:
                raise ValueError("Got more than one target task")
            has_target = True
    if not has_target:
        raise ValueError("Needs at least one target task")

    return settings


def settings_from_file(config_path):
    """Loads and parses a parameter file.

    Parameters
    ===========
    config_path : str
        The path to the parameter file, formatted as json.

    Returns
    ===========
    settings : dict, A dictionary with the parameters
    """

    try:
        with open(config_path, 'r') as f:
            p = json.loads(json_minify(f.read()))
    except Exception as e:
        raise ValueError(
            "Couldn't read config file: %s. Exception: %s" % (config_path, str(e)))

    # add default values for missing settings:
    with open(DEFAULTPATH, 'r') as f:
        defaults = json.loads(json_minify(f.read()))

    settings = Settings(
        utils.recursive_merge(
            # merge defaults
            utils.recursive_merge(p, defaults),
            # ultimately overwrite settings from environ vars of the form PIE_{var}
            parse_env_settings(defaults), overwrite=True))

    # store the config path too:
    settings.config_path = config_path

    if settings.verbose:
        print("\n::: Loaded Config :::\n")
        print(yaml.dump(dict(settings)))

    return check_settings(merge_task_defaults(settings))
